"""
Implementation of the matching_anchors_filter script
"""
from __future__ import annotations

import argparse
import logging
import typing
from dataclasses import dataclass
from pathlib import Path
from typing import Any, List, Optional, Sequence, TypedDict, Union, cast

import configpile as cp
import numpy as np
import tybles as tb
from filelock import FileLock
from numpy.typing import NDArray
from typing_extensions import Annotated

from ..lib.data import LoggingLevel, NameRow, PathPattern, PickleProtocol
from ..lib.io import open_pickle, save_pickle
from ..lib.math import local_max, make_continuum
from ..lib.util import log_task_name_and_time
from ..rassine.data import RassineBasicOutput, RassineParameters, RassinePickle


@dataclass(frozen=True)
class MatchingAnchorsRow:
    """
    Describes the effect of the matching_anchors step
    """

    name: str
    is_master: bool
    n_anchors_before: int
    n_anchors_ooc: int
    n_anchors_twin: int
    n_anchors_adding: int

    @staticmethod
    def schema() -> tb.Schema[MatchingAnchorsRow]:
        """
        Returns the tybles Schema for this dataclass

        :meta private:
        """
        return tb.schema(
            MatchingAnchorsRow,
            order_columns=True,
            missing_columns="error",
            extra_columns="drop",
        )


class MasterToolPickle(TypedDict):
    """Format of the master tool created by the matching_anchors_scan step"""

    curve: NDArray[np.bool_]
    border: NDArray[np.float64]
    wave: NDArray[np.float64]
    threshold: float
    tolerance: float
    fraction: float
    nb_copies_master: int

    #: Warning: only the final path component of the master file is included
    master_filename: Optional[str]


class AnchorParameters(TypedDict):
    #: Master tool file used to compute the matching anchors
    master_tool: str
    master_filename: Optional[str]
    #: Parameters used to define the clusters considered has significant
    threshold: float
    #: Parameters used to define the clusters considered has significant
    tolerance: float
    fraction: float
    #: Number of copies of the master spectra. Such parameter is used to fix the position of the anchors points on the master spectra by overweighting it in the timeseries
    nb_copies_master: int
    #: Continuum called. Usually performed on the 'output' continua. Used later by YARARA to construct the tree of the reduction.
    sub_dico_used: str


class AnchorOutput(TypedDict):
    #: Rassine parameters used to match the anchor points locations
    parameters: AnchorParameters
    #: Rassine fitted continuum
    continuum_linear: NDArray[np.float64]
    #: Wavelength values of the anchors points used (continuum_linear being a linear interpolation between them)
    anchor_wave: NDArray[np.float64]
    #: Flux values of the anchors points used (continuum_linear being a linear interpolation between them)
    anchor_flux: NDArray[np.float64]
    #: Index values of the anchors points used (continuum_linear being a linear interpolation between them)
    anchor_index: NDArray[np.int64]


class AnchorPickle(TypedDict):
    #: Wavelength of the spectrum
    wave: NDArray[np.float64]
    #: Raw flux of the spectrum in photon count units
    flux: NDArray[np.float64]
    #: Raw flux uncertainties of the spectrum in photon count units
    flux_err: Optional[NDArray[np.float64]]
    #: Smoothed raw spectrum by RASSINE used to fit the continuum. Avoid to fit the upper envelop of the noise.
    flux_used: NDArray[np.float64]
    #: Rassine output continuum (individual spectra normalisation)
    output: RassineBasicOutput
    #: Rassine derived parameters
    parameters: RassineParameters
    #: Rassine output continuum (spectra time-series cluster identified)
    matching_anchors: AnchorOutput


class MatchingDiffParameters(TypedDict):
    reference_continuum: str
    savgol_window: int
    recenter: bool
    sub_dico_used: str  # always matching_nachors


class MatchingDiffOutput(TypedDict):
    parameters: MatchingDiffParameters
    continuum_linear: NDArray[np.float64]


class MatchingPickle(TypedDict):
    #: Wavelength of the spectrum
    wave: NDArray[np.float64]
    #: Raw flux of the spectrum in photon count units
    flux: NDArray[np.float64]
    #: Raw flux uncertainties of the spectrum in photon count units
    flux_err: Optional[NDArray[np.float64]]
    #: RASSINE smoothed flux of the spectrum used to fit the continuum (in order to avoid fitting the upper envelop of the noise).
    flux_used: NDArray[np.float64]
    #: Rassine output continuum (individual spectra normalisation)
    output: RassineBasicOutput
    #: Rassine derived parameters
    parameters: RassineParameters
    #: Rassine output continuum (spectra time-series cluster identified)
    matching_anchors: AnchorOutput
    #: Rassine output continuum (spectra time-series cluster identified + low-pass filtering)
    matching_diff: MatchingDiffOutput



import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Sequence, cast

import configpile as cp
import numpy as np
from numpy.typing import NDArray
from typing_extensions import Annotated, Literal

from ..lib.io import open_pickle
from ..matching.data import MatchingPickle


@dataclass(frozen=True)
class Task(cp.Config):

    #
    # Common information
    #

    prog_ = Path(__file__).stem
    ini_strict_sections_ = [Path(__file__).stem]
    ini_relaxed_sections_ = [Path(__file__).stem.split("_")[0]]
    env_prefix_ = "RASSINE"

    #
    # Task specific information
    #

    #: First pickle to compare
    input1: Annotated[
        Path,
        cp.Param.store(cp.parsers.path_parser, positional=cp.Positional.ONCE, long_flag_name=None),
    ]

    #: Second pickle to compare
    input2: Annotated[
        Path,
        cp.Param.store(cp.parsers.path_parser, positional=cp.Positional.ONCE, long_flag_name=None),
    ]

    #: Kind of output to compare
    kind: Annotated[
        Literal["output", "matching_anchors", "matching_diff"],
        cp.Param.store(cp.Parser.from_choices(["output", "matching_anchors", "matching_diff"])),
    ]

    #: Threshold for acceptance in normalized units
    threshold: Annotated[float, cp.Param.store(cp.parsers.float_parser, default_value="0.005")]


def get_continuum_linear(
    data: MatchingPickle, kind: Literal["output", "matching_anchors", "matching_diff"]
) -> NDArray[np.float64]:
    """Returns the flux data corresponding to the given key

    Args:
        data: RASSINE output data
        kind: Kind of spectrum to get
    """
    if kind == "output":
        return data["output"]["continuum_linear"]
    elif kind == "matching_anchors":
        return data["matching_anchors"]["continuum_linear"]
    elif kind == "matching_diff":
        return data["matching_diff"]["continuum_linear"]


def compute_distance(
    data1: MatchingPickle,
    data2: MatchingPickle,
    kind: Literal["output", "matching_anchors", "matching_diff"],
) -> np.float64:
    """Returns the distance between two outputs

    Args:
        data1: RASSINE output data (1st)
        data2: RASSINE output data (2nd)
        kind: Kind of output used for comparison
    """
    return np.std(
        data1["flux"] / get_continuum_linear(data1, kind)
        - data2["flux"] / get_continuum_linear(data2, kind)
    )


def run(t: Task):
    data1 = cast(MatchingPickle, open_pickle(t.input1))
    data2 = cast(MatchingPickle, open_pickle(t.input2))
    distance = compute_distance(data1, data2, t.kind)
    if distance <= t.threshold:
        print(f"Distance below threshold: {distance} <= {t.threshold}")
        sys.exit(0)
    else:
        print(f"Distance above threshold: {distance} > {t.threshold}")
        sys.exit(1)


def cli():
    run(Task.from_command_line_())
